@@ -28,6 +28,7 @@
  ** Feb, 2011 - Some debugging code for checking Generic file format parsing
  ** Nov, 2011 - Some additional fixed to deal with fixed width fields for strings in dataset rows
  ** Sept 4, 2017 - change gzFile * to gzFile
+ ** August 26, 2021 - Handling fixed width strings of length 0, Better handling for situations where logical ordering and physical ordering of data groups do not agree
  **
  *************************************************************/
 
@@ -179,6 +180,7 @@ static int fread_ASTRING_fw(ASTRING *destination, FILE *instream, int length){
     }
   } else {
     destination->value = 0;
+    fseek(instream, length, SEEK_CUR);
   }
   return 1;
 }
@@ -225,6 +227,7 @@ static int fread_AWSTRING_fw(AWSTRING *destination, FILE *instream, int length){
     }
   } else {
     destination->value = 0;
+    fseek(instream, length, SEEK_CUR);
   }
   
   return 1;
@@ -929,6 +932,7 @@ static int gzread_ASTRING_fw(ASTRING *destination, gzFile instream, int length){
     }
   } else {
     destination->value = 0;
+    gzseek(instream, length, SEEK_CUR);
   }
   return 1;
 }
@@ -976,6 +980,7 @@ static int gzread_AWSTRING_fw(AWSTRING *destination, gzFile instream, int length
 
   } else {
     destination->value = 0;
+    gzseek(instream, length, SEEK_CUR);
   }
   
   return 1;
@@ -1478,6 +1483,7 @@ SEXP Read_Generic(SEXP filename){
       fseek(infile, my_data_set.file_pos_last, SEEK_SET);
       Free_generic_data_set(&my_data_set);
     }
+    fseek(infile, my_data_group.file_position_nextgroup, SEEK_SET);
     Free_generic_data_group(&my_data_group);
   }
   Free_generic_data_header(&my_data_header);
@@ -1687,8 +1693,8 @@ static SEXP decode_nvt_triplet(nvt_triplet triplet){
 
 static SEXP data_header_R_List(generic_data_header *my_data_header){
 
-  SEXP return_value, return_names;	
-  SEXP tmp_sexp, tmp_names;
+  SEXP return_value= R_NilValue, return_names= R_NilValue;	
+  SEXP tmp_sexp= R_NilValue, tmp_names= R_NilValue;
   char *temp;
   int i;
 
@@ -1778,10 +1784,123 @@ static SEXP data_header_R_List(generic_data_header *my_data_header){
 
 
 
+
+
+static SEXP data_header_R_List_full(generic_data_header *my_data_header){
+
+  SEXP return_value= R_NilValue, return_names= R_NilValue;	
+  SEXP tmp_sexp= R_NilValue, tmp_names= R_NilValue, tmp_type= R_NilValue, tmp_value= R_NilValue;
+  char *temp;
+  int i;
+
+  PROTECT(return_value = allocVector(VECSXP,8));	
+
+  PROTECT(tmp_sexp= allocVector(STRSXP,1));
+  if (my_data_header->data_type_id.len > 0){
+    SET_STRING_ELT(tmp_sexp,0,mkChar(my_data_header->data_type_id.value));
+  }
+  SET_VECTOR_ELT(return_value,0,tmp_sexp);
+  UNPROTECT(1); 	
+
+  PROTECT(tmp_sexp= allocVector(STRSXP,1));  
+  if (my_data_header->unique_file_id.len > 0){
+    SET_STRING_ELT(tmp_sexp,0,mkChar(my_data_header->unique_file_id.value));
+  }
+  SET_VECTOR_ELT(return_value,1,tmp_sexp);
+  UNPROTECT(1); 	
+
+  PROTECT(tmp_sexp= allocVector(STRSXP,1));
+  if (my_data_header->Date_time.len > 0){
+    temp = Calloc(my_data_header->Date_time.len+1,char);
+    wcstombs(temp, my_data_header->Date_time.value, my_data_header->Date_time.len);
+    SET_STRING_ELT(tmp_sexp,0,mkChar(temp));  
+    Free(temp);
+  }
+  SET_VECTOR_ELT(return_value,2,tmp_sexp);
+  UNPROTECT(1); 
+ 
+  PROTECT(tmp_sexp= allocVector(STRSXP,1));
+  if (my_data_header->locale.len > 0){
+    temp = Calloc(my_data_header->locale.len+1,char);
+    wcstombs(temp, my_data_header->locale.value, my_data_header->locale.len);
+    SET_STRING_ELT(tmp_sexp,0,mkChar(temp));  
+    Free(temp);
+  }
+  SET_VECTOR_ELT(return_value,3,tmp_sexp);
+  UNPROTECT(1); 
+   
+  PROTECT(tmp_sexp= allocVector(INTSXP,1));
+  INTEGER(tmp_sexp)[0] =  (int32_t)my_data_header->n_name_type_value;
+  SET_VECTOR_ELT(return_value,4,tmp_sexp);
+  UNPROTECT(1); 
+
+  PROTECT(tmp_sexp= allocVector(VECSXP,3));
+  PROTECT(tmp_value = allocVector(VECSXP,  my_data_header->n_name_type_value));
+  PROTECT(tmp_names = allocVector(STRSXP, my_data_header->n_name_type_value));
+  PROTECT(tmp_type = allocVector(STRSXP, my_data_header->n_name_type_value));
+  
+  for (i=0; i < my_data_header->n_name_type_value; i++){
+     SET_VECTOR_ELT(tmp_value,i,decode_nvt_triplet(my_data_header->name_type_value[i]));
+     temp = Calloc(my_data_header->name_type_value[i].name.len+1,char);
+     wcstombs(temp, my_data_header->name_type_value[i].name.value, my_data_header->name_type_value[i].name.len);
+     SET_STRING_ELT(tmp_names,i,mkChar(temp));
+     Free(temp);
+     temp = Calloc(my_data_header->name_type_value[i].type.len+1,char);
+     wcstombs(temp, my_data_header->name_type_value[i].type.value, my_data_header->name_type_value[i].type.len);
+     SET_STRING_ELT(tmp_type,i,mkChar(temp));
+     Free(temp);
+
+     
+  } 
+  setAttrib(tmp_value, R_NamesSymbol, tmp_names);
+  SET_VECTOR_ELT(tmp_sexp,0,tmp_names);
+  SET_VECTOR_ELT(tmp_sexp,1,tmp_value);
+  SET_VECTOR_ELT(tmp_sexp,2,tmp_type);
+
+  PROTECT(return_names = allocVector(STRSXP,3));
+  SET_STRING_ELT(return_names,0,mkChar("Name"));
+  SET_STRING_ELT(return_names,1,mkChar("Value"));
+  SET_STRING_ELT(return_names,2,mkChar("Type"));
+  setAttrib(tmp_sexp, R_NamesSymbol, return_names); 
+  UNPROTECT(1);
+
+  SET_VECTOR_ELT(return_value,5,tmp_sexp);
+  UNPROTECT(4); 
+
+  PROTECT(tmp_sexp= allocVector(INTSXP,1));
+  INTEGER(tmp_sexp)[0] =  (int32_t)my_data_header->n_parent_headers;
+  SET_VECTOR_ELT(return_value,6,tmp_sexp);
+  UNPROTECT(1); 
+  
+  PROTECT(tmp_sexp= allocVector(VECSXP,my_data_header->n_parent_headers)); 
+  if (my_data_header->n_parent_headers > 0){
+   for (i =0; i < my_data_header->n_parent_headers; i++){
+      SET_VECTOR_ELT(tmp_sexp,i,data_header_R_List_full(my_data_header->parent_headers[i]));
+    }
+  }
+  SET_VECTOR_ELT(return_value,7,tmp_sexp);
+  UNPROTECT(1); 
+
+
+  PROTECT(return_names = allocVector(STRSXP,8));
+  SET_STRING_ELT(return_names,0,mkChar("DataTypeID"));
+  SET_STRING_ELT(return_names,1,mkChar("UniqueFileID"));
+  SET_STRING_ELT(return_names,2,mkChar("DateTime"));
+  SET_STRING_ELT(return_names,3,mkChar("Locale"));
+  SET_STRING_ELT(return_names,4,mkChar("NumberOfNameValueType"));
+  SET_STRING_ELT(return_names,5,mkChar("NVTList"));
+  SET_STRING_ELT(return_names,6,mkChar("NumberOfParentHeaders"));
+  SET_STRING_ELT(return_names,7,mkChar("ParentHeaders"));
+  setAttrib(return_value, R_NamesSymbol, return_names); 
+  UNPROTECT(2);
+  return return_value;
+}
+
+
 static SEXP data_group_R_list(generic_data_group *my_data_group){
 
   SEXP return_value;
-  SEXP tmp_sexp=R_NilValue, return_names;
+  SEXP tmp_sexp=R_NilValue, return_names= R_NilValue;
   char *temp;
 
   PROTECT(return_value =  allocVector(VECSXP,2));
@@ -1810,8 +1929,8 @@ static SEXP data_group_R_list(generic_data_group *my_data_group){
 
 static SEXP generic_data_set_R_List(generic_data_set *my_data_set){
 
-  SEXP return_value, return_names;
-  SEXP tmp_sexp, tmp_names;
+  SEXP return_value= R_NilValue, return_names= R_NilValue;
+  SEXP tmp_sexp= R_NilValue, tmp_names= R_NilValue;
   int i;
   char *temp;
 
@@ -1867,6 +1986,99 @@ static SEXP generic_data_set_R_List(generic_data_set *my_data_set){
 
 
 
+static SEXP generic_data_set_R_List_full(generic_data_set *my_data_set){
+
+  SEXP return_value = R_NilValue, return_names = R_NilValue;
+  SEXP tmp_sexp = R_NilValue, tmp_names = R_NilValue, tmp_type = R_NilValue, tmp_value = R_NilValue, tmp_size = R_NilValue;
+  int i;
+  char *temp;
+
+  PROTECT(return_value =  allocVector(VECSXP,4));
+  
+  PROTECT(tmp_sexp= allocVector(STRSXP,1));  
+  if (my_data_set->data_set_name.len > 0){
+    temp = Calloc(my_data_set->data_set_name.len+1,char);
+    wcstombs(temp, my_data_set->data_set_name.value, my_data_set->data_set_name.len);
+    SET_STRING_ELT(tmp_sexp,0,mkChar(temp));  
+    Free(temp);
+  }
+  SET_VECTOR_ELT(return_value,0,tmp_sexp);
+  UNPROTECT(1);
+
+  PROTECT(tmp_sexp= allocVector(VECSXP,3));
+  PROTECT(tmp_names =  allocVector(STRSXP,my_data_set->n_name_type_value));
+  PROTECT(tmp_type =  allocVector(STRSXP,my_data_set->n_name_type_value));
+  PROTECT(tmp_value =  allocVector(VECSXP,my_data_set->n_name_type_value));
+  for (i=0; i < my_data_set->n_name_type_value; i++){
+    SET_VECTOR_ELT(tmp_value,i,decode_nvt_triplet(my_data_set->name_type_value[i]));
+    temp = Calloc(my_data_set->name_type_value[i].name.len+1,char);
+    wcstombs(temp, my_data_set->name_type_value[i].name.value, my_data_set->name_type_value[i].name.len);
+    SET_STRING_ELT(tmp_names,i,mkChar(temp));
+    Free(temp);
+    temp = Calloc(my_data_set->name_type_value[i].type.len+1,char);
+    wcstombs(temp, my_data_set->name_type_value[i].type.value, my_data_set->name_type_value[i].type.len);
+    SET_STRING_ELT(tmp_type,i,mkChar(temp));
+    Free(temp);
+  } 
+  setAttrib(tmp_value, R_NamesSymbol, tmp_names);
+  SET_VECTOR_ELT(tmp_sexp,0,tmp_names);
+  SET_VECTOR_ELT(tmp_sexp,1,tmp_value);
+  SET_VECTOR_ELT(tmp_sexp,2,tmp_type);
+  SET_VECTOR_ELT(return_value,1,tmp_sexp);
+  UNPROTECT(4); 
+
+  PROTECT(tmp_sexp= allocVector(VECSXP,my_data_set->ncols));  
+  SET_VECTOR_ELT(return_value,2,tmp_sexp);
+  PROTECT(tmp_names =  allocVector(STRSXP,my_data_set->ncols));
+  for (i=0; i < my_data_set->ncols; i++){
+     temp = Calloc(my_data_set->col_name_type_value[i].name.len+1,char);
+     wcstombs(temp, my_data_set->col_name_type_value[i].name.value, my_data_set->col_name_type_value[i].name.len);
+     SET_STRING_ELT(tmp_names,i,mkChar(temp));
+     Free(temp);
+  }
+  setAttrib(tmp_sexp, R_NamesSymbol, tmp_names); 
+  UNPROTECT(2);
+
+
+  PROTECT(tmp_sexp = allocVector(VECSXP,3));
+  PROTECT(tmp_names =  allocVector(STRSXP,my_data_set->ncols));
+  PROTECT(tmp_value =  allocVector(INTSXP,my_data_set->ncols));
+  PROTECT(tmp_size =  allocVector(INTSXP,my_data_set->ncols));
+  for (i=0; i < my_data_set->ncols; i++){
+    temp = Calloc(my_data_set->col_name_type_value[i].name.len+1,char);
+    wcstombs(temp, my_data_set->col_name_type_value[i].name.value, my_data_set->col_name_type_value[i].name.len);
+    SET_STRING_ELT(tmp_names,i,mkChar(temp));
+    Free(temp);
+    INTEGER(tmp_value)[i] = (int) my_data_set->col_name_type_value[i].type;
+    INTEGER(tmp_size)[i] = (int) my_data_set->col_name_type_value[i].size;
+  }
+  SET_VECTOR_ELT(tmp_sexp,0,tmp_names);
+  SET_VECTOR_ELT(tmp_sexp,1,tmp_value);
+  SET_VECTOR_ELT(tmp_sexp,2,tmp_size);
+
+  PROTECT(return_names = allocVector(STRSXP,3));
+  SET_STRING_ELT(return_names,0,mkChar("Name"));
+  SET_STRING_ELT(return_names,1,mkChar("ValueType"));
+  SET_STRING_ELT(return_names,2,mkChar("Size"));
+  setAttrib(tmp_sexp, R_NamesSymbol, return_names); 
+  UNPROTECT(1);
+
+  
+  SET_VECTOR_ELT(return_value,3,tmp_sexp);
+  UNPROTECT(4);
+  
+
+  PROTECT(return_names = allocVector(STRSXP,4));
+  SET_STRING_ELT(return_names,0,mkChar("Name"));
+  SET_STRING_ELT(return_names,1,mkChar("NVTList"));
+  SET_STRING_ELT(return_names,2,mkChar("DataColumns"));
+  SET_STRING_ELT(return_names,3,mkChar("DataColumnNTS"));
+  setAttrib(return_value, R_NamesSymbol, return_names); 
+  UNPROTECT(2);
+  return return_value;
+
+}
+
 
 
 
@@ -1876,7 +2088,7 @@ static SEXP generic_data_set_R_List(generic_data_set *my_data_set){
 
 static SEXP generic_data_set_rows_R_List(generic_data_set *data_set, int col){
 
-  SEXP return_value;
+  SEXP return_value= R_NilValue;
   int i,j;
   char *temp;  
 
@@ -1935,8 +2147,10 @@ static SEXP generic_data_set_rows_R_List(generic_data_set *data_set, int col){
   case 7:	
      PROTECT(return_value = allocVector(STRSXP, data_set->nrows));
      for (i=0; i < data_set->nrows; i++){
-	temp = (char *)((ASTRING *)data_set->Data[j])[i].value;
-	SET_STRING_ELT(return_value,i,mkChar(temp));
+       if ((int32_t)((ASTRING *)data_set->Data[j])[i].len > 0){
+	 temp = (char *)((ASTRING *)data_set->Data[j])[i].value;
+	 SET_STRING_ELT(return_value,i,mkChar(temp));
+       }
      }
      break;
   case 8:	
@@ -1956,10 +2170,12 @@ static SEXP generic_data_set_rows_R_List(generic_data_set *data_set, int col){
 
 
 
-SEXP Read_Generic_R_List(SEXP filename){
+SEXP Read_Generic_R_List(SEXP filename, SEXP reducedoutput){
 
   int i,j,k;
 
+  int shorten_NVT = asInteger(reducedoutput);
+  
   SEXP return_value = R_NilValue;
   SEXP return_names;
   SEXP temp_sxp = R_NilValue,temp_sxp2 = R_NilValue,temp_names = R_NilValue,temp_names2 = R_NilValue;	
@@ -1995,8 +2211,13 @@ SEXP Read_Generic_R_List(SEXP filename){
   SET_VECTOR_ELT(return_value,0,file_header_R_List(&my_header));
 
   /* Data Header is Second Element of Return List */
-  SET_VECTOR_ELT(return_value,1,data_header_R_List(&my_data_header));
+  if (shorten_NVT){
+    SET_VECTOR_ELT(return_value,1,data_header_R_List(&my_data_header));
+  } else {
+    SET_VECTOR_ELT(return_value,1,data_header_R_List_full(&my_data_header));
+  }
 
+    
   /* Data Groups are it Third Element of Return List */	
   /* Now Read Data groups */	
   
@@ -2015,8 +2236,14 @@ SEXP Read_Generic_R_List(SEXP filename){
     
     PROTECT(temp_names2 = allocVector(STRSXP,my_data_group.n_data_sets));	
     for (j=0; j < my_data_group.n_data_sets; j++){
-      read_generic_data_set(&my_data_set,infile); 
-      temp_sxp2 = generic_data_set_R_List(&my_data_set);
+      read_generic_data_set(&my_data_set,infile);
+
+      if (shorten_NVT){
+	temp_sxp2 = generic_data_set_R_List(&my_data_set);
+      } else {
+	temp_sxp2 =  generic_data_set_R_List_full(&my_data_set);
+      }
+      
       SET_VECTOR_ELT(VECTOR_ELT(VECTOR_ELT(temp_sxp,k),1),j,temp_sxp2);
 
       temp = Calloc(my_data_set.data_set_name.len+1,char);
@@ -2033,6 +2260,7 @@ SEXP Read_Generic_R_List(SEXP filename){
       fseek(infile, my_data_set.file_pos_last, SEEK_SET);
       Free_generic_data_set(&my_data_set);
     }
+    fseek(infile, my_data_group.file_position_nextgroup, SEEK_SET);
     setAttrib(VECTOR_ELT(VECTOR_ELT(temp_sxp,k),1), R_NamesSymbol, temp_names2); 
     UNPROTECT(1);	
 
